"""
MathApp class for creating periodic and timed callbacks.
"""

from ggame.mathapp import MathApp, _MathDynamic


class Timer(_MathDynamic):
    """
    The Timer class instantiates an object whose basic function is to report
    the number of seconds since its creation by calling the object as a
    function with an empty argument list.

    The Timer class accepts no arguments during creation.

    Example of use:

    .. literalinclude:: ../examples/timertimer.py

    """

    def __init__(self):
        super().__init__()
        self._once = []
        self._callbacks = {}
        self._time = 0
        self.reset()
        MathApp.addDynamic(self)  # always dynamically defined

    def reset(self):
        """
        Set the reference time to the MathApp current time. If the timer is reset
        before the app initializes then do nothing.

        :returns: None
        """
        self._reset = MathApp.time

    @property
    def time(self):
        """
        Attribute is always updated with the number of seconds since the
        timer was created.
        """
        return self._time

    @time.setter
    def time(self, value):
        pass

    def step(self):
        nexttimers = []
        calllist = []
        self._time = MathApp.time - self._reset
        while self._once and self._once[0][0] <= MathApp.time:
            tickinfo = self._once.pop(0)
            if tickinfo[1]:  # periodic?
                nexttimers.append(
                    (tickinfo[1], self._callbacks[tickinfo][0])
                )  # delay, callback
            calllist.append(
                self._callbacks[tickinfo].pop(0)
            )  # remove callback and queue it
            if not self._callbacks[tickinfo]:  # if the callback list is empty
                del self._callbacks[tickinfo]  # remove the dictionary entry altogether
        for tickadd in nexttimers:
            self.callAfter(tickadd[0], tickadd[1], True)  # keep it going
        for call in calllist:
            call(self)

    def callAfter(self, delay, callback, periodic=False):
        """
        Set a callback to occur either once or periodically. The callback
        function should accept a single parameter which will be a reference
        to the timer object that called it.

        :param float delay: The number of seconds to wait before executing
            the callback function
        :param function callback: The callback function to call on timer
            expiration
        :param boolean periodic: Set True if the callback function should
            be executed periodically
        :returns: None
        """
        key = (self._time + delay, delay if periodic else 0)
        self._once.append(key)
        callbacklist = self._callbacks.get(key, [])
        callbacklist.append(callback)
        self._callbacks[key] = callbacklist
        self._once.sort()

    def callAt(self, calltime, callback):
        """
        Set a callback to occur at a specific time (seconds since
        the Timer object was created or since its
        :func:`~ggame.timer.Timer.reset` method was called.

        :param float time: The time to wait since timer creation or reset
            before calling the callback function.
        :param function callback: The callback function to call
        :returns: None
        """
        self.callAfter(calltime - self._time, callback)

    def callEvery(self, period, callback):
        """
        Set a callback to occur periodically. The callback
        function should accept a single parameter which will be a reference
        to the timer object that called it.

        :param float period: The number of seconds to wait before each
            execution of the callback function
        :param function callback: The callback function to call on timer
            expiration
        :returns: None
        """
        self.callAfter(period, callback, True)

    def __call__(self):
        return self._time



"""
These MathApp-based digital logic classes are experimental.
"""

from abc import ABCMeta, abstractmethod
from ggame.mathapp import _MathDynamic


# decorator for _getvalue or any value handler that may experience recursion
def _recursiontrap(handler):
    def trapmagic(self):
        """
        An attempt to catch recursion (doesn't work).
        """
        if not self.ingetvalue:
            self.ingetvalue = True
            self.lastget = handler(self)
            self.ingetvalue = False
            return self.lastget
        self.ingetvalue = False
        return self.lastget

    return trapmagic


class _BoolDevice(_MathDynamic, metaclass=ABCMeta):
    """
    Base class for boolean objects.

    :Required Arguments:

    :param int mininputqty: The minimum number of inputs possible.

    :Optional Keyword Arguments:

        * **namedinputs**  (*list[str]*) List of input names.
    """

    def __init__(self, mininputqty, **kwargs):
        self.inp = [None] * mininputqty
        self.enable = True
        namedinputs = kwargs.get("namedinputs", [])
        self._indict = {name: self.eval(None) for name in namedinputs}
        self.ingetvalue = False
        self.lastget = False
        self.resetval = False
        self.firsttime = True
        self._enable = None
        self._input = None
        super().__init__()

    @property
    def inp(self):
        """
        Report the list of input references.
        """
        return self._input

    @inp.setter
    def inp(self, val):
        try:
            self._input = [self.eval(v) for v in list(val)]
        except TypeError:
            self._input = [self.eval(val)]

    # Enable attribute controls the "tri-state" of output
    @property
    def enable(self):
        """
        Report the enable state of the object.
        """
        return self._enable

    @enable.setter
    def enable(self, val):
        self._enable = self.eval(val)

    @abstractmethod
    @_recursiontrap  # MUST use with any implementation that may recurse!
    def _getvalue(self):
        return None

    @staticmethod
    def _inputState(value):
        """
        interprets a value that could be single input or a list of inputs!
        """
        try:
            inputs = [] + value
        except TypeError:
            inputs = [value]
        scalars = [v() for v in inputs]
        ones = scalars.count(True) + scalars.count(1)
        zeros = scalars.count(False) + scalars.count(0)
        if ones > 0 and zeros > 0:
            raise ValueError("Conflicting inputs found")
        if ones > 0:
            return True
        if zeros > 0:
            return False
        return None

    def __call__(self):
        if self.enable:
            return self._getvalue()
        return None

    def getinput(self, inputname):
        """
        Retrieve input by name.

        :param str inputname: Name to look up.
        """
        return self._inputState(self._indict[inputname])

    def setinput(self, inputname, reference):
        """
        Set an input connection.

        :param str inputname: Name to assign.
        :param function reference: Callable object or function connected to input.
        """
        self._indict[inputname] = self.eval(reference)


class _BoolOneInput(_BoolDevice):
    """
    Base class for one-input boolean objects. No required inputs.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(1, *args, **kwargs)

    @abstractmethod
    @_recursiontrap  # MUST use with any implementation that may recurse!
    def _getvalue(self):
        return None


class _BoolMultiInput(_BoolDevice):
    """
    Base class for multiple-input boolean objects. No required inputs.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(2, *args, **kwargs)

    @abstractmethod
    @_recursiontrap  # MUST use with any implementation that may recurse!
    def _getvalue(self):
        return None


class BoolNOT(_BoolOneInput):
    """
    Logical NOT boolean gate.
    """

    @_recursiontrap
    def _getvalue(self):
        inval = self._inputState(self.inp[0])
        if inval is None:
            return True  # equivalent to an "open" input
        return not inval


class BoolAND(_BoolMultiInput):
    """
    Logical AND boolean gate. Multiple inputs.
    """

    @_recursiontrap
    def _getvalue(self):
        for v in self._input:
            if not self._inputState(v):
                return False
        return True


class BoolNOR(_BoolMultiInput):
    """
    Logical NOR boolean gate. Multiple inputs.
    """

    @_recursiontrap
    def _getvalue(self):
        for v in self._input:
            if self._inputState(v):
                return False
        return True


class BoolNAND(_BoolMultiInput):
    """
    Logical NAND boolean gate. Multiple inputs.
    """

    @_recursiontrap
    def _getvalue(self):
        for v in self._input:
            if not self._inputState(v):
                return True
        return False


class BoolSRFF(_BoolOneInput):
    """
    Logical Set/Reset (SR) FlipFlop.

    :Optional Keyword Arguments:

    :param class gateclass: One of BoolNAND or BoolNOR (default).
    """

    def __init__(self, *args, **kwargs):
        kwargs["namedinputs"] = ["R", "S"]
        super().__init__(*args, **kwargs)
        gate = kwargs.get("gateclass", BoolNOR)
        self.ic1 = gate()
        self.ic2 = gate()

    # we can only assign IC1 and IC2 inputs when this device's inputs are set
    def setinput(self, inputname, reference):
        super().setinput(inputname, reference)
        if inputname == "R":
            self.ic1.In = reference, self.ic2
        elif inputname == "S":
            self.ic2.In = reference, self.ic1

    def _getvalue(self):
        return self.ic1()

    # pylint: disable=invalid-name
    def q_(self):
        """
        Report value of Q_ output.
        """
        return self.ic2()

    # pylint: disable=invalid-name
    def q(self):
        """
        Report value of Q output.
        """
        return self._getvalue()



"""
ggame events are objects that are created by the ggame system as a result of
some user action (mouse, keyboard). If the ggame application has called
:meth:`~ggame.app.App.listenKeyEvent` or :meth:`~ggame.app.App.listenMouseEvent` then
an appropriate Event object is instantiated by ggame and returned to a
user-provided handler function. The ggame user code should examine the
attributes of the Event object to find out more information about the event that
occurred.
"""
import traceback


class _Event:
    def __init__(self, hwevent):
        self.hwevent = hwevent
        self.type = hwevent.type
        """String representing the type of received system event."""
        self.consumed = False
        """
        Set the `consumed` member of the event to prevent the event
        from being received by any more handler methods.
        """

    def route(self, evtlist):
        """
        Execute all callbacks configured for this event.
        """
        for callback in reversed(evtlist):
            if not self.consumed:
                try:
                    callback(self)
                except BaseException:
                    traceback.print_exc()
                    raise


class MouseEvent(_Event):
    """
    A MouseEvent object encapsulates information about a user mouse
    action that is being reported by the system.  This class is not instantiated
    by the ggame user.
    """

    mousemove = "mousemove"
    mousedown = "mousedown"
    mouseup = "mouseup"
    click = "click"
    dblclick = "dblclick"
    mousewheel = "wheel"

    def __init__(self, app, hwevent):
        """
        The event is initialized by the system, with a `hwevent` input parameter.
        """
        super().__init__(hwevent)
        self.wheeldelta = 0
        """Integer representing up/down motion of the scroll wheel."""
        if self.type == self.mousewheel:
            self.wheeldelta = hwevent.deltaY
        else:
            self.wheeldelta = 0
        rect = app.win.renderer.view.getBoundingClientRect()
        xscale = app.win.width / rect.width
        yscale = app.win.height / rect.height
        self.x = (hwevent.clientX - rect.left) * xscale
        """The window x-coordinate of the mouse pointer when the event occurred."""
        self.y = (hwevent.clientY - rect.top) * yscale
        """The window y-coordinate of the mouse pointer when the event occurred."""


class KeyEvent(_Event):
    """
    A KeyEvent object encapsulates information regarding a user keyboard
    action that is being reported by the system. This class is not instantiated
    by the ggame user.
    """

    no_location = 0
    right_location = 2
    left_location = 1
    keydown = "keydown"
    keyup = "keyup"
    keypress = "keypress"
    keys = {
        8: "backspace",
        9: "tab",
        13: "enter",
        16: "shift",
        17: "ctrl",
        18: "alt",
        19: "pause/break",
        20: "caps lock",
        27: "escape",
        32: "space",
        33: "page up",
        34: "page down",
        35: "end",
        36: "home",
        37: "left arrow",
        38: "up arrow",
        39: "right arrow",
        40: "down arrow",
        45: "insert",
        46: "delete",
        48: "0",
        49: "1",
        50: "2",
        51: "3",
        52: "4",
        53: "5",
        54: "6",
        55: "7",
        56: "8",
        57: "9",
        65: "a",
        66: "b",
        67: "c",
        68: "d",
        69: "e",
        70: "f",
        71: "g",
        72: "h",
        73: "i",
        74: "j",
        75: "k",
        76: "l",
        77: "m",
        78: "n",
        79: "o",
        80: "p",
        81: "q",
        82: "r",
        83: "s",
        84: "t",
        85: "u",
        86: "v",
        87: "w",
        88: "x",
        89: "y",
        90: "z",
        91: "left window key",
        92: "right window key",
        93: "select key",
        96: "numpad 0",
        97: "numpad 1",
        98: "numpad 2",
        99: "numpad 3",
        100: "numpad 4",
        101: "numpad 5",
        102: "numpad 6",
        103: "numpad 7",
        104: "numpad 8",
        105: "numpad 9",
        106: "multiply",
        107: "add",
        109: "subtract",
        110: "decimal point",
        111: "divide",
        112: "f1",
        113: "f2",
        114: "f3",
        115: "f4",
        116: "f5",
        117: "f6",
        118: "f7",
        119: "f8",
        120: "f9",
        121: "f10",
        122: "f11",
        123: "f12",
        144: "num lock",
        145: "scroll lock",
        186: "semicolon",
        187: "equal sign",
        188: "comma",
        189: "dash",
        190: "period",
        191: "forward slash",
        192: "grave accent",
        219: "open bracket",
        220: "back slash",
        221: "close bracket",
        222: "single quote",
    }
    """Dictionary mapping key code integers to textual key description."""

    def __init__(self, hwevent):
        """
        The event is initialized by the system, with a `hwevent` input parameter.
        """
        super().__init__(hwevent)
        self.keynum = hwevent.keyCode
        """The `keynum` attribute identifies a keycode (number)."""
        self.key = self.keys[hwevent.keyCode]
        """
        The `key` attribute identifes the key in text form (e.g. 'back slash').

        The list of key numbers and description strings follows::

            8: 'backspace',
            9: 'tab',
            13: 'enter',
            16: 'shift',
            17: 'ctrl',
            18: 'alt',
            19: 'pause/break',
            20: 'caps lock',
            27: 'escape',
            32: 'space',
            33: 'page up',
            34: 'page down',
            35: 'end',
            36: 'home',
            37: 'left arrow',
            38: 'up arrow',
            39: 'right arrow',
            40: 'down arrow',
            45: 'insert',
            46: 'delete',
            48: '0',
            49: '1',
            50: '2',
            51: '3',
            52: '4',
            53: '5',
            54: '6',
            55: '7',
            56: '8',
            57: '9',
            65: 'a',
            66: 'b',
            67: 'c',
            68: 'd',
            69: 'e',
            70: 'f',
            71: 'g',
            72: 'h',
            73: 'i',
            74: 'j',
            75: 'k',
            76: 'l',
            77: 'm',
            78: 'n',
            79: 'o',
            80: 'p',
            81: 'q',
            82: 'r',
            83: 's',
            84: 't',
            85: 'u',
            86: 'v',
            87: 'w',
            88: 'x',
            89: 'y',
            90: 'z',
            91: 'left window key',
            92: 'right window key',
            93: 'select key',
            96: 'numpad 0',
            97: 'numpad 1',
            98: 'numpad 2',
            99: 'numpad 3',
            100: 'numpad 4',
            101: 'numpad 5',
            102: 'numpad 6',
            103: 'numpad 7',
            104: 'numpad 8',
            105: 'numpad 9',
            106: 'multiply',
            107: 'add',
            109: 'subtract',
            110: 'decimal point',
            111: 'divide',
            112: 'f1',
            113: 'f2',
            114: 'f3',
            115: 'f4',
            116: 'f5',
            117: 'f6',
            118: 'f7',
            119: 'f8',
            120: 'f9',
            121: 'f10',
            122: 'f11',
            123: 'f12',
            144: 'num lock',
            145: 'scroll lock',
            186: 'semicolon',
            187: 'equal sign',
            188: 'comma',
            189: 'dash',
            190: 'period',
            191: 'forward slash',
            192: 'grave accent',
            219: 'open bracket',
            220: 'back slash',
            221: 'close bracket',
            222: 'single quote'

        """



"""
Tools for loading and playing sound resources in ggame applications.
"""

from ggame.sysdeps import SND_Sound


class SoundAsset:
    """
    Class representing a single sound asset (sound file, such as .mp3 or .wav).

    :param str url: The URL or file name of the desired sound. Sound file
        formats may include `.wav` or `.mp3`, subject to browser compatibility.

    :returns: The asset instance
    """

    def __init__(self, url):
        self.url = url
        """
        A string containing the url or name of the asset file.
        """


class Sound(object):  # pylint: disable=useless-object-inheritance
    """
    The Sound class represents a sound, with methods for controlling
    when and how the sound is played in the application.

    :param SoundAsset asset: A valid :class:`SoundAsset` instance.

    :returns: the Sound instance
    """

    def __init__(self, asset):
        self._asset = asset
        """
        A reference to the sound asset instance.
        """
        self._snd = SND_Sound(self._asset.url)
        """
        A reference to the underlying sound object provided by the system.
        """
        self._snd.load()

    def play(self):
        """
        Play the sound once.
        """
        self.stop()
        self._snd.play()

    def loop(self):
        """
        Play the sound continuously, looping forever.
        """
        self.stop()
        self._snd.loop()
        self._snd.play()

    def stop(self):
        """
        Stop playing the sound.
        """
        self._snd.stop()

    @property
    def volume(self):
        """
        The volume property is a number ranging from 0-100 that
        represents the volume or intensity of the sound when it is playing.
        """
        return self._snd.getVolume()

    @volume.setter
    def volume(self, value):
        self._snd.setVolume(value)



"""
MathApp label classes for displaying text.
"""

from ggame.mathapp import _MathVisual
from ggame.asset import TextAsset


class Label(_MathVisual):
    """
    Create a text label on the screen. This is a
    subclass of :class:`~ggame.sprite.Sprite` and
    :class:`~ggame.mathapp._MathVisual` but most of the inherited
    members are of little use and are not shown in the documentation.

    :param \\*args:
        See below
    :param \\**kwargs:
        See below

    :Required Arguments:
        * **pos**  (*tuple(float,float)*) Screen position of the label, which may
            be a literal tuple of floats, or a reference to any object or
            function that returns or evaluates to a tuple of floats.
        * **text** (*str*) Text to appear in the label. This may be a literal
            string or a reference to any object or function that returns or
            evaluates to a string.

    :Optional Keyword Arguments:
        * **positioning** (*str*) One of 'logical' or 'physical'
        * **size** (*int*) Size of text font (in pixels)
        * **width** (*int*) Width of the label (in pixels)
        * **color** (*Color*) Valid :class:`~ggame.asset.Color` object

    Example:

    .. literalinclude:: ../examples/labellabel.py

    """

    _posinputsdef = ["pos"]
    _nonposinputsdef = ["text"]

    def __init__(self, *args, **kwargs):
        """
        Required Inputs

        * **pos** position of label
        * **text** text contents of label
        """
        super().__init__(TextAsset(""), *args, **kwargs)
        self.touchAsset()

    def _buildAsset(self):
        return TextAsset(
            self._nposinputs.text(),  # pylint: disable=no-member
            style=f"{self._stdinputs.size()}px Courier",
            width=self._stdinputs.width(),
            fill=self._stdinputs.color(),
        )

    def __call__(self):
        return self._nposinputs.text()  # pylint: disable=no-member

    def physicalPointTouching(self, ppos):
        _ppos = self._spposinputs.pos
        return (
            ppos[0] >= _ppos[0]
            and ppos[0] <= _ppos[0] + self._sstdinputs.width
            and ppos[1] >= _ppos[1]
            and ppos[1] <= _ppos[1] + self._sstdinputs.size
        )

    def translate(self, pdisp):
        pass



"""
ggame assets and related objects (:class:`Color` and :class:`LineStyle`) are
classes that encapsulate and represent displayable images. A single asset
may be used in multiple sprites. Animated assets may be created from any
image that includes multiple images within it (i.e. a sprite sheet).
"""

from ggame.sysdeps import (
    GFX_Rectangle,
    GFX_Texture,
    GFX_Graphics,
    GFX_Texture_fromImage,
    GFX_Text,
)


class Frame(object):  # pylint: disable=useless-object-inheritance
    """
    Frame is a utility class for expressing the idea of a rectangular region.

    Initializing parameters describe the position of the upper-left corner
    of the frame, and the frame's width and height. The units are *typically*
    in pixels, though a frame can be generic.

    :param int x: X-coordinate of frame upper-left corner
    :param int y: Y-coordinate of frame upper-left corner
    :param int w: Width of the frame
    :param int h: Height of the frame
    """

    def __init__(self, x, y, w, h):
        self.gfx = GFX_Rectangle(x, y, w, h)
        """
        `_gfx` is a reference to the underlying object provided by the system.
        """
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    @property
    def x(self):
        """
        X-coordinate of the upper left hand corner of this frame.
        """
        return self.gfx.x

    @x.setter
    def x(self, value):
        self.gfx.x = value

    @property
    def y(self):
        """
        Y-coordinate of the upper left hand corner of this frame.
        """
        return self.gfx.y

    @y.setter
    def y(self, value):
        self.gfx.y = value

    @property
    def w(self):
        """
        Width of the frame.
        """
        return self.gfx.width

    @w.setter
    def w(self, value):
        self.gfx.width = value

    @property
    def h(self):
        """
        Height of the frame.
        """
        return self.gfx.height

    @h.setter
    def h(self, value):
        self.gfx.height = value

    @property
    def center(self):
        """
        The `center` property computes a coordinate pair (tuple) for the
        center of the frame.

        The `center` property, when set, redefines the `x` and `y` properties
        of the frame in order to make the center agree with the coordinates
        (tuple) assigned to it.
        """

        return (self.x + self.w // 2, self.y + self.h // 2)

    @center.setter
    def center(self, value):
        c = self.center
        self.x += value[0] - c[0]
        self.y += value[1] - c[1]


class _Asset(object):  # pylint: disable=useless-object-inheritance
    """
    Base class for all game asset objects.

    The `ggame.Asset` class is set up to understand the concept
    of multiple instances of an asset. This is currently only used for
    image-based assets.
    """

    def __init__(self):
        self.gfxlist = [None]
        """
        A list of the underlying system objects used to represent this
        asset.
        """

    @property
    def gfx(self):
        """
        `gfx` property represents the underlying system object used to represent
        this asset. If this asset is composed of multiple assets, then the
        **first** asset is referenced by `gfx`.
        """
        return self.gfxlist[0]

    @gfx.setter
    def gfx(self, value):
        self.gfxlist[0] = value

    def __len__(self):
        return len(self.gfxlist)

    def __getitem__(self, key):
        return self.gfxlist[key]

    def __setitem__(self, key, value):
        self.gfxlist[key] = value

    def __iter__(self):
        class Iter:
            """
            Iterator for accessing discrete images within an asset.
            """

            def __init__(self, image):
                self.obj = image
                self.n = len(image.gfxlist)
                self.i = 0

            def __iter__(self):
                return self

            def __next__(self):
                if self.i == self.n:
                    raise StopIteration
                self.i += 1
                return self.obj.gfxlist[self.i]

        return Iter(self)

    def destroy(self):
        """
        Destroy or deallocate any underlying graphics resources used by the
        asset. Call this method on any asset that is no longer being used.
        """
        if hasattr(self, "gfx"):
            try:
                for gfx in self.gfxlist:
                    try:
                        gfx.destroy(True)
                    except BaseException:
                        pass
            except BaseException:
                pass


class ImageAsset(_Asset):
    """
    The ImageAsset class connects ggame to a specific image **file**.

    :param str url: All ImageAsset instances must specify a file name or
        url where a jpg or png image is located.

    :param Frame frame: If the desired sprite image exists in only a smaller
        sub-section of the original image, then specify an area `within` the
        image using `frame` parameter, which must be a valid :class:`Frame`
        instance.

    :param int qty=0: If the image file actually is a *collection* of images,
        such as a so-called *sprite sheet*, then the ImageAsset class supports
        defining a list of images, provided they exist in the original image
        as a **row** of evenly spaced images or a **column** of images. To
        specify this, provide the `qty` (quantity) of images in the row or
        column.

    :param str direction='horizontal': For an image *sprite sheet*, specify
        whether the images are oriented in a `'vertical'` or `'horizontal'`
        arrangement.

    :param int margin=0: If there is a gap between successive images in an image
        *sprite sheet*, then specify the size of the gap (in pixels). When
        used in this way, the `frame` parameter must define the
        area of only the **first** image in the collection; all subsequent
        images in the list are assumed to be the same size, but separated by the
        `margin` value.

    Example:

    .. literalinclude:: ../examples/assetimage.py
    """

    def __init__(self, url, frame=None, qty=1, direction="horizontal", margin=0):
        super().__init__()
        self.url = url
        """
        A string that represents the path or url of the original file.
        """
        del self.gfxlist[0]
        self.width = self.height = 0
        self.append(url, frame, qty, direction, margin)

    def append(self, url, frame=None, qty=1, direction="horizontal", margin=0):
        """
        Append a texture asset from a new image file (or url). This method
        allows you to build a collection of images into an asset (such as you
        might need for an animated sprite), but without using a single
        sprite sheet image.

        The parameters for the `append` method are identical to those
        supplied to the :class:`ImageAsset` initialization method.

        This method allows you to build up an asset that consists of
        multiple rows or columns of images in a sprite sheet or sheets.
        """
        gfx = GFX_Texture_fromImage(url, False)
        dx = 0
        dy = 0
        for i in range(qty):
            if frame is not None:
                self.width = frame.w
                self.height = frame.h
                if direction == "horizontal":
                    dx = frame.w + margin
                elif direction == "vertical":
                    dy = frame.h + margin
                f = Frame(frame.x + dx * i, frame.y + dy * i, frame.w, frame.h)
                gfx = GFX_Texture(gfx, f.gfx)
            else:
                self.width = gfx.width
                self.height = gfx.height
            self.gfxlist.append(gfx)


class Color:
    """
    The Color class is used to represent colors and/or colors with
    transparency.

    :param int color: an integer in the conventional format (usually
        as a hexadecimal literal, e.g. 0xffbb33 that represents
        the three color components, red, green and blue)

    :param float alpha: a transparency value, or `alpha` as a floating-point
        number in the range of 0.0 to 1.0 where 0.0 represents
        completely transparent and 1.0 represents completely opaque.

    Example:

    .. literalinclude:: ../examples/assetcolor.py
    """

    _colornames = {0xFFFFFF: "WHITE", 0x000000: "BLACK"}

    def __init__(self, color, alpha):
        self.color = color
        self.alpha = alpha
        self.name = f"Color(0x{int(self.color):06X}, {self.alpha})"
        if alpha == 1.0:
            self.name = self._colornames.get(self.color, self.name)

    def __eq__(self, other):
        return (
            isinstance(self, type(other))
            and self.color == other.color
            and self.alpha == other.alpha
        )

    def __repr__(self):
        return self.name


BLACK = Color(0x000000, 1.0)
"""
Default black color
"""
WHITE = Color(0xFFFFFF, 1.0)
"""
Default white color
"""


class LineStyle:
    """
    The LineStyle class is used to represent line style when
    drawing geometrical objects such as rectangles, ellipses, etc.

    :param int width: the `width` of the line in pixels

    :param Color color: the `color` as a valid :class:`Color` instance.

    Example:

    .. literalinclude:: ../examples/assetlinestyle.py

    This defines a 3-pixel wide green line.
    """

    def __init__(self, width, color):
        """ """
        self.width = width
        self.color = color

    def __eq__(self, other):
        return (
            isinstance(self, type(other))
            and self.width == other.width
            and self.color == other.color
        )

    def __repr__(self):
        return f"LineStyle({self.width}, {self.color})"


BLACKLINE = LineStyle(1, BLACK)
"""
Default thin black line
"""
WHITELINE = LineStyle(1, WHITE)
"""
Default thin white line
"""


class _GraphicsAsset(_Asset):
    def __init__(self):
        super().__init__()
        GFX_Graphics.clear()


class _CurveAsset(_GraphicsAsset):
    def __init__(self, line):
        super().__init__()
        GFX_Graphics.lineStyle(line.width, line.color.color, line.color.alpha)


class _ShapeAsset(_CurveAsset):
    def __init__(self, line, fill):
        super().__init__(line)
        GFX_Graphics.beginFill(fill.color, fill.alpha)


class RectangleAsset(_ShapeAsset):
    """
    The RectangleAsset is a "virtual" asset that is created on the
    fly without requiring creation of an image file.

    :param int width: Rectangle width, in pixels
    :param int height: Rectangle height, in pixels
    :param LineStyle line=BLACKLINE: The color and width of the rectangle border
    :param Color fill=BLACK: The color of the rectangle body
    """

    def __init__(self, width, height, line=BLACKLINE, fill=BLACK):
        super().__init__(line, fill)
        self.width = width
        self.height = height
        self.gfx = GFX_Graphics.drawRect(0, 0, self.width, self.height).clone()
        """The `gfx` property represents the underlying system object."""
        self.gfx.visible = False


class CircleAsset(_ShapeAsset):
    """
    The `ggame.CircleAsset` is a "virtual" asset that is created on the
    fly without requiring creation of an image file.

    :param int radius: Circle radius, in pixels
    :param LineStyle line=BLACKLINE: The color and width of the circle border
    :param Color fill=BLACK: The color of the circle body
    """

    def __init__(self, radius, line=BLACKLINE, fill=BLACK):
        super().__init__(line, fill)
        self.radius = radius
        self.gfx = GFX_Graphics.drawCircle(0, 0, self.radius).clone()
        """The `gfx` property represents the underlying system object."""
        self.gfx.visible = False


class EllipseAsset(_ShapeAsset):
    """
    The `ggame.EllipseAsset` is a "virtual" asset that is created on the
    fly without requiring creation of an image file.

    :param int halfw: Ellipse semi-axis dimension in the horizontal direction
        (half the width of the ellipse), in pixels
    :param int halfh: Ellipse semi-axis dimension in the vertical direction
        (half the height of the ellipse), in pixels
    :param LineStyle line=BLACKLINE: The color and width of the ellipse border
    :param Color fill=BLACK: The color of the ellipse body

    """

    def __init__(self, halfw, halfh, line=BLACKLINE, fill=BLACK):
        super().__init__(line, fill)
        self.halfw = halfw
        self.halfh = halfh
        self.gfx = GFX_Graphics.drawEllipse(0, 0, self.halfw, self.halfh).clone()
        """The `gfx` property represents the underlying system object."""
        self.gfx.visible = False


class PolygonAsset(_ShapeAsset):
    """
    The PolygonAsset is a "virtual" asset that is created on the
    fly without requiring creation of an image file.

    Note: you should not specificy absolute screen coordinates for this
    asset, since you will use the :class:`Sprite` position to locate your
    polygon on the screen.

    :param list path: A list of pixel-coordinate tuples. These coordinates
        should not be in absolute screen coordinates, but should be relative to
        the desired 'center' of the resulting :class:`Sprite`. The final
        coordinate pair in the list must be the same as the first.
    :param LineStyle line=BLACKLINE: The color and width of the ellipse border
    :param Color fill=BLACK: The color of the ellipse body

    Example:

    .. literalinclude:: ../examples/assetpolygon.py

    """

    def __init__(self, path, line=BLACKLINE, fill=BLACK):
        super().__init__(line, fill)
        self.path = path[:]
        jpath = []
        # close the path if necessary
        if self.path[0] != self.path[-1]:
            self.path.append(self.path[0])
        for point in self.path:
            jpath.extend(point)
        self.gfx = GFX_Graphics.drawPolygon(jpath).clone()
        """The `gfx` property represents the underlying system object."""
        self.gfx.visible = False


class LineAsset(_CurveAsset):
    """
    The LineAsset is a "virtual" asset that is created on the
    fly without requiring creation of an image file. A LineAsset instance
    represents a single line segment.

    Note: you should not specificy absolute screen coordinates for this
    asset, since you will use the :class:`Sprite` position to locate your
    line on the screen. The line segment will begin at pixel coordinates (0,0),
    and will end at the (x,y) coordinates given below.

    As the LineAsset does not cover a region, only a :class:`LineStyle`
    argument must be supplied (`line`) to specify the color.

    :param int x: x-coordinate of the line endpoint, in pixel units
    :param int y: y-coordinate of the line endpoint, in pixel units
    :param LineStyle line=BLACKLINE: The color and width of the ellipse border
    """

    def __init__(self, x, y, line=BLACKLINE):
        super().__init__(line)
        self.delta_x = x
        """
        This attribute represents the `x` parameter supplied during
        instantiation.
        """
        self.delta_y = y
        """
        This attribute represents the `y` parameter supplied during
        instantiation.
        """
        GFX_Graphics.moveTo(0, 0)
        self.gfx = GFX_Graphics.lineTo(self.delta_x, self.delta_y).clone()
        """The `gfx` property represents the underlying system object."""
        self.gfx.visible = False


class TextAsset(_GraphicsAsset):
    """
    The TextAsset is a "virtual" asset that is created on the fly
    without requiring creation of an image file. A TextAsset instance
    represents a block of text, together with its styling (font, color, etc.).

    :param str text: The text that should be displayed

    :param \\**kwargs: Optional formatting and style attributes (below)

    :param str style='20px Arial': Text style, size and typeface. Example::

            'italic 20pt Helvetica'

    :keyword int width=100: Width of the text block on screen, in pixels.
    :param Color fill=BLACK: :class:`Color` instance to specify color
        and transparency of the text.
    :param str align='left': Alignment style of the block. One of 'left',
        'center', or 'right'.

    Full example:

    .. literalinclude:: ../examples/assettext.py

    """

    def __init__(self, text, **kwargs):
        """ """
        super().__init__()
        self.text = text
        self.style = kwargs.get("style", "20px Arial")
        width = kwargs.get("width", 100)
        self.fill = kwargs.get("fill", Color(0, 1))
        self.align = kwargs.get("align", "left")
        self.gfx = GFX_Text(
            self.text,
            {
                "font": self.style,
                "fill": self.fill.color,
                "align": self.align,
                "wordWrap": True,
                "wordWrapWidth": width,
            },
        )
        self.gfx.alpha = self.fill.alpha
        self.gfx.visible = False

    def clone(self):
        """
        Create a duplicate asset with the current style settings.

        :returns: A text asset that is identical to the original, but with current
            styles.
        :rtype: TextAsset
        """
        return type(self)(
            self.text,
            style=self.style,
            width=self.width,
            fill=self.fill,
            align=self.align,
        )

    @property
    def width(self):
        """
        Width of the rendered text asset in pixels
        """
        return self.gfx.width

    @property
    def height(self):
        """
        Height of the rendered text asset in pixels
        """
        return self.gfx.height



"""
MathApp input classes for accepting user numeric and pushbutton input.
"""

from ggame.label import Label
from ggame.asset import TextAsset
from ggame.mathapp import MathApp


class InputNumeric(Label):
    """
    Create a :class:`~ggame.label.Label` that can be selected to accept
    new **numeric** input from the keyboard. This is a
    subclass of :class:`~ggame.sprite.Sprite` and
    :class:`~ggame.mathapp._MathVisual` but most of the inherited
    members are of little use and are not shown in the documentation.

    :param tuple(float,float) pos: Screen position of the control, which may
            be a literal tuple of floats, or a reference to any object or
            function that returns or evaluates to a tuple of floats.

    :param float val: Initial value of the input text

    :param \\**kwargs:
        See below

    :Optional Keyword Arguments:
        * **fmt** (*str*) a Python format string (default is {0:.2f})
        * **positioning** (*str*) One of 'logical' or 'physical'
        * **size** (*int*) Size of text font (in pixels)
        * **width** (*int*) Width of the label (in pixels)
        * **color** (*Color*) Valid :class:`~ggame.asset.Color` object

    Example:

    .. literalinclude:: ../examples/inputinputnumeric.py

    """

    def __init__(self, pos, val, **kwargs):
        self._fmt = kwargs.get("fmt", "{0:.2f}")
        self._val = self.eval(val)()  # initialize to simple numeric
        self._savedval = self._val
        self._text = ""
        self._updateText()
        super().__init__(pos, self._textValue, **kwargs)
        self.selectable = True

    def _textValue(self):
        return self._text()

    def _updateText(self, editing=False):
        if editing:
            self._text = self.eval(f"{self._val}")
        else:
            self._text = self.eval(self._fmt.format(self._val))

    def processEvent(self, event):
        if event.key in "0123456789insertdelete":
            key = event.key
            if event.key == "insert":
                key = "-"
            elif event.key == "delete":
                key = "."
            if self._text() == "0":
                self._text = self.eval("")
            self._text = self.eval(self._text() + key)
            self.touchAsset()
        elif event.key in ["enter", "escape"]:
            if event.key == "enter":
                try:
                    self._val = float(self._text())
                except ValueError:
                    self._val = self._savedval
                self._savedval = self._val
            self.unselect()

    def select(self):
        super().select()
        self._savedval = self._val
        self._val = 0
        self._updateText(True)
        self.touchAsset()
        MathApp.listenKeyEvent("keypress", "*", self.processEvent)

    def unselect(self):
        super().unselect()
        self._val = self._savedval
        self._updateText()
        self.touchAsset()
        try:
            MathApp.unlistenKeyEvent("keypress", "*", self.processEvent)
        except ValueError:
            pass

    def __call__(self):
        return self._val


class InputButton(Label):
    """
    Create a :class:`~ggame.label.Label` that can be clicked with a mouse
    to execute a user-defined function. This is a
    subclass of :class:`~ggame.sprite.Sprite` and
    :class:`~ggame.mathapp._MathVisual` but most of the inherited
    members are of little use and are not shown in the documentation.

    :param function callback: A reference to a function to execute, passing
        this button object, when the button is clicked
    :param tuple(float,float) pos: Screen position of the control, which may
            be a literal tuple of floats, or a reference to any object or
            function that returns or evaluates to a tuple of floats.

    :param float val: Initial value of the input text

    :param \\*args:
        See below
    :param \\**kwargs:
        See below

    :Required Arguments:
        * **pos**  (*tuple(float,float)*) Screen position of the button, which may
            be a literal tuple of floats, or a reference to any object or
            function that returns or evaluates to a tuple of floats.
        * **text** (*str*) Text to appear in the button. This may be a literal
            string or a reference to any object or function that returns or
            evaluates to a string.


    :Optional Keyword Arguments:
        * **positioning** (*str*) One of 'logical' or 'physical'
        * **size** (*int*) Size of text font (in pixels)
        * **width** (*int*) Width of the label (in pixels)
        * **color** (*Color*) Valid :class:`~ggame.asset.Color` object

    Example:

    .. literalinclude:: ../examples/inputinputbutton.py

    """

    def __init__(self, callback, *args, **kwargs):
        """
        Required Inputs

        * **pos** position of button
        * **text** text of button
        * **callback** reference of a function to execute, passing this button object
        """
        super().__init__(*args, **kwargs)
        self.touchAsset()
        self._callback = callback
        self.selectable = True

    def _buildAsset(self):
        return TextAsset(
            self._nposinputs.text(),  # pylint: disable=no-member
            style=f"bold {self._stdinputs.size()}px Courier",
            width=self._stdinputs.width(),
            fill=self._stdinputs.color(),
        )

    def select(self):
        super().select()
        if self._callback:
            self._callback(self)
        self.unselect()



"""
ggame extensions for modeling spacecraft in planetary orbit
"""

from math import pi, degrees, radians, atan2, sin, cos, sqrt
from ggame.asset import LineStyle, Color
from ggame.mathapp import MathApp
from ggame.circle import Circle
from ggame.point import ImagePoint
from ggame.timer import Timer
from ggame.label import Label


class Planet(MathApp):
    """
    Initialize the Planet object.

    Optional keyword parameters are supported:

    :param float viewscale: Pixels per meter in graphics display. Default is 10.
    :param float radius: Radius of the planet in meters. Default is Earth radius.
    :param float planetmass: Mass of the planet in kg. Default is Earth mass.
    :param int color: Color of the planet. Default is greenish (0x008040).
    :param float viewalt: Altitude of initial viewpoint in meters. Default is rocket
      altitude.
    :param float viewanom: True anomaly (angle) of initial viewpoint in radians.
      Default is the rocket anomaly.
    :param float viewanomd: True anomaly (angle) of initial viewpoing in degrees.
      Default is the rocket anomaly.

    Example:

    .. literalinclude:: ../examples/astroplanet.py

    """

    def __init__(self, **kwargs):
        scale = kwargs.get("viewscale", 10)  # 10 pixels per meter default
        self.radius = kwargs.get("radius", 6.371e6)  # Earth - meters
        self.mass = kwargs.get("planetmass", 5.9722e24)  # Earth - kg
        self.color = kwargs.get("color", 0x008040)  # greenish
        self.viewaltitude = kwargs.get("viewalt", 0)  # how high to look
        self.viewanomaly = kwargs.get("viewanom", pi / 2)  # where to look
        self.viewanomalyd = kwargs.get("viewanomd", degrees(self.viewanomaly))
        self._planetcircle = Circle(
            (0, 0),
            self.radius,
            style=LineStyle(1, Color(self.color, 1)),
            color=Color(self.color, 0.5),
        )
        super().__init__(scale)

    def run(self, userfunc=None):
        """
        Execute the Planet (and Rocket) simulation without setting the initial view.
        """
        self.runWithRocket()

    def runWithRocket(self, rocket=None):
        """
        Execute the Planet (and Rocket) simulation.

        :Optional parameters:

        :param Rocket rocket: Reference to a Rocket object - sets the initial view
        :returns: None
        """
        if rocket:
            self.viewaltitude = rocket.altitude
            self.viewanomaly = rocket.tanomaly
        r = self.radius + self.viewaltitude
        self.view_position = (r * cos(self.viewanomaly), r * sin(self.viewanomaly))
        super().run()


class Rocket(ImagePoint):
    """
    Rocket is a class for simulating the motion of a projectile through space,
    acted upon by arbitrary forces (thrust) and by gravitaitonal
    attraction to a single planetary object.

    Required parameters:

    :param Planet planet:  Reference to a :class:`Planet` object.

    Optional keyword parameters are supported:

    :param str bitmap:  Url of a suitable bitmap image for the rocket (png
        recommended), default is `images/rocket.png`
    :param float bitmapscale:  Scale factor for bitmap. Default is 0.1
    :param float velocity:  Initial rocket speed. Default is zero.
    :param float directiond:  Initial rocket direction in degrees. Default is zero.
    :param float direction:  Initial rocket direction in radians. Default is zero.
    :param float tanomalyd:  Initial rocket true anomaly in degrees. Default is 90.
    :param float tanomaly:  Initial rocket true anomaly in radians. Default is pi/2.
    :param float altitude:  Initial rocket altitude in meters. Default is zero.
    :param bool showstatus:  Boolean displays flight parameters on screen. Default
        is True.
    :param (int, int) statuspos:  Tuple with screen x,y coordinates of flight
        parameters. Default is upper left.
    :param list[str] statuslist: List of status names to include in flight parameters.
      Default is all, consisting of: "velocity", "acceleration", "course",
      "altitude", "thrust", "mass", "trueanomaly", "scale", "timezoom",
      "shiptime"
    :param str leftkey: A :class:`ggame.event.KeyEvent` key identifier that will serve
        as the "rotate left" key while controlling the ship. Default is 'left arrow'.
    :param str rightkey: A :class:`ggame.event.KeyEvent` key identifier that will serve
        as the "rotate right" key while controlling the ship. Default is 'right arrow'.

    Following parameters may be set as a constant value, or pass in the
    name of a function that will return the value dynamically or the
    name of a `ggmath` UI control that will return the value.

    :param function or float timezoom:  Scale factor for time zoom. Factor =
        10^timezoom
    :param function or float heading:  Direction to point the rocket in (must be
        radians)
    :param fucntion or float mass: Mass of the rocket (must be kg)
    :param function or float thrust: Thrust of the rocket (must be N)

    Animation related parameters may be ignored if no sprite animation:

    :param Frame bitmapframe:  ((x1,y1),(x2,y2)) tuple defines a region in the bitmap
    :param int bitmapqty: Number of bitmaps -- used for animation effects
    :param str bitmapdir: "horizontal" or "vertical" use with animation effects. Default
        is "horizontal"
    :param int bitmapmargin: Pixels between successive animation frames
    :param float tickrate: Frequency of spacecraft dynamics calculations (Hz)

    Example:

    .. literalinclude:: ../examples/astrorocket.py

    """

    def __init__(self, planet, **kwargs):
        self._xy = (1000000, 1000000)
        self.planet = planet
        """Reference to an app object of :class:`Planet` class"""
        self._bmurl = kwargs.get("bitmap", self.getImagePath("rocket.png"))
        """URL of a bitmap image to use for the rocket"""
        self._bitmapframe = kwargs.get("bitmapframe", None)
        """:class:`ggame.asset.Frame` that specifies location of displayed
            image in bitmap image
        """
        self._bitmapqty = kwargs.get("bitmapqty", 1)
        """Number of sub-images in the bitmap image"""
        self._bitmapdir = kwargs.get("bitmapdir", "horizontal")
        """Orientation of the sub-images for animation (one of "horizontal" or
            "vertical")
        """
        self._bitmapmargin = kwargs.get("bitmapmargin", 0)
        """Margin between the sub-images for animation."""
        self.tickrate = kwargs.get("tickrate", 30)
        """Target animation frames per second."""
        # status display
        statusfuncs = [
            self._velocityText,
            self._accelerationText,
            self._courseDegreesText,
            self._altitudeText,
            self._thrustText,
            self._massText,
            self._trueAnomalyDegreesText,
            self._scaleText,
            self._timeZoomText,
            self._shipTimeText,
        ]
        statuslist = [
            "velocity",
            "acceleration",
            "course",
            "altitude",
            "thrust",
            "mass",
            "trueanomaly",
            "scale",
            "timezoom",
            "shiptime",
        ]

        self._showstatus = kwargs.get("showstatus", True)
        """Boolean: True to show stats on screen, else False."""
        self._statuspos = kwargs.get("statuspos", [10, 10])
        """Location on screen for showing the stats."""
        self._statusselect = kwargs.get("statuslist", statuslist)
        """Reference to a list of statuses to display."""
        self.localheading = 0
        """Heading (angle) of travel"""
        # dynamic parameters
        self.timezoom = self.eval(kwargs.get("timezoom", 0))
        """Reference to a function that will dynamically determine the
        time zoom factor. An integer: 1,2,3 faster, -1, slower.
        """
        self.heading = self.eval(kwargs.get("heading", self._getheading))
        """Reference to a function that will dynamically calculate spaceship
        orientation or heading (in radians)
        """
        self.mass = self.eval(kwargs.get("mass", 1))
        """Reference to a function that will dynamically calculate the
        spaceship mass (in kg).
        """
        self.thrust = self.eval(kwargs.get("thrust", 0))
        """Reference to a funtion that will dynamically calculate the rocket
        thrust (in N).
        """
        # end dynamic
        super().__init__(
            self._bmurl,
            self._getposition,
            frame=self._bitmapframe,
            qty=self._bitmapqty,
            direction=self._bitmapdir,
            margin=self._bitmapmargin,
        )
        self.scale = kwargs.get("bitmapscale", 0.1)  # small
        initvel = kwargs.get("velocity", 0)  # initial velocity
        initdird = kwargs.get("directiond", 0)  # initial direction, degrees
        initdir = kwargs.get("direction", radians(initdird))
        tanomaly = kwargs.get("tanomaly", pi / 2)  # position angle
        tanomaly = radians(kwargs.get("tanomalyd", degrees(tanomaly)))
        altitude = kwargs.get("altitude", 0)  #
        r = altitude + self.planet.radius
        self._xy = (r * cos(tanomaly), r * sin(tanomaly))
        # default heading control if none provided by user
        leftkey = kwargs.get("leftkey", "left arrow")
        rightkey = kwargs.get("rightkey", "right arrow")
        if self.heading == self._getheading:  # pylint: disable=comparison-with-callable
            Planet.listenKeyEvent("keydown", leftkey, self._turn)
            Planet.listenKeyEvent("keydown", rightkey, self._turn)
        self._timer = Timer()
        self.shiptime = 0  # track time on shipboard
        self._timer.callEvery(1 / self.tickrate, self._dynamics)
        self._lasttime = self._timer.time
        self._v_vect = [initvel * cos(initdir), initvel * sin(initdir)]
        self._a_vect = [0, 0]
        # keep track of on-screen resources
        self._labels = []
        # set up status display
        if self._showstatus:
            self.addStatusReport(statuslist, statusfuncs, self._statusselect)

    @staticmethod
    def _vadd(v1, v2):
        """
        Vector add utility.
        """
        return [v1[i] + v2[i] for i in (0, 1)]

    @staticmethod
    def _vmul(s, v):
        """
        Scalar vector multiplication utility.
        """
        return [s * v[i] for i in (0, 1)]

    @staticmethod
    def _vmag(v):
        """
        Vector magnitude function.
        """
        return sqrt(v[0] ** 2 + v[1] ** 2)

    def _turn(self, event):
        """

        Respond to left/right turning key events.
        """
        increment = pi / 50 * (1 if event.key == "left arrow" else -1)
        self.localheading += increment

    def _getposition(self):
        return self._xy

    # override or define externally!
    def _getheading(self):
        """
        User override function for dynamically determining the heading.

        :returns: float
        """
        return self.localheading

    # functions available for reporting flight parameters to UI
    def _velocityText(self):
        """
        Report the velocity in m/s as a text string.
        """
        return f"Velocity:     {self.velocity:8.1f} m/s"

    def _accelerationText(self):
        """
        Report the acceleration in m/s as a text string.
        """
        return f"Acceleration: {self.acceleration:8.1f} m/s²"

    def _courseDegreesText(self):
        """
        Report the heading in degrees (zero to the right) as a text string.
        """
        return f"Course:       {degrees(atan2(self._v_vect[1], self._v_vect[0])):8.1f}°"

    def _thrustText(self):
        """
        Report the thrust level in Newtons as a text string.
        """
        return f"Thrust:       {self.thrust():8.1f} N"

    def _massText(self):
        """
        Report the spacecraft mass in kilograms as a text string.
        """
        return f"Mass:         {self.mass():8.1f} kg"

    def _trueAnomalyDegreesText(self):
        """
        Report the true anomaly in degrees as a text string.
        """
        return f"True Anomaly: {self.tanomalyd:8.1f}°"

    def _trueAnomalyRadiansText(self):
        """
        Report the true anomaly in radians as a text string.
        """
        return f"True Anomaly: {self.tanomaly:8.4f}"

    def _altitudeText(self):
        """
        Report the altitude in meters as a text string.
        """
        return f"Altitude:     {self.altitude:8.1f} m"

    def _radiusText(self):
        """
        Report the radius (distance to planet center) in meters as a text string.
        """
        return f"Radius:       {self.r:8.1f} m"

    def _scaleText(self):
        """
        Report the view scale (pixels/meter) as a text string.
        """
        return f"View Scale:   {self.planet.scale:8.6f} px/m"

    def _timeZoomText(self):
        """
        Report the time acceleration as a text string.
        """
        return f"Time Zoom:    {float(self.timezoom()):8.1f}"

    def _shipTimeText(self):
        """
        Report the elapsed time as a text string.
        """
        return f"Elapsed Time: {float(self.shiptime):8.1f} s"

    def _dynamics(self, timer):
        """
        Perform one iteration of the simulation using runge-kutta 4th order method.
        """
        # set time duration equal to time since last execution
        tick = 10 ** self.timezoom() * (timer.time - self._lasttime)
        self.shiptime = self.shiptime + tick
        self._lasttime = timer.time
        # 4th order runge-kutta method
        # (https://sites.temple.edu/math5061/files/2016/12/final_project.pdf)
        # and http://spiff.rit.edu/richmond/nbody/OrbitRungeKutta4.pdf  (succinct,
        # but with a typo)
        self._a_vect = k1v = self._ar(self._xy)
        k1r = self._v_vect
        k2v = self._ar(self._vadd(self._xy, self._vmul(tick / 2, k1r)))
        k2r = self._vadd(self._v_vect, self._vmul(tick / 2, k1v))
        k3v = self._ar(self._vadd(self._xy, self._vmul(tick / 2, k2r)))
        k3r = self._vadd(self._v_vect, self._vmul(tick / 2, k2v))
        k4v = self._ar(self._vadd(self._xy, self._vmul(tick, k3r)))
        k4r = self._vadd(self._v_vect, self._vmul(tick, k3v))
        self._v_vect = [
            self._v_vect[i] + tick / 6 * (k1v[i] + 2 * k2v[i] + 2 * k3v[i] + k4v[i])
            for i in (0, 1)
        ]
        self._xy = [
            self._xy[i] + tick / 6 * (k1r[i] + 2 * k2r[i] + 2 * k3r[i] + k4r[i])
            for i in (0, 1)
        ]
        if self.altitude < 0:
            self._v_vect = [0, 0]
            self._a_vect = [0, 0]
            self.altitude = 0

    # generic force as a function of position
    def _fr(self, pos):
        """
        Compute the net force vector on the rocket, as a function of the
        position vector.
        """
        self.rotation = self.heading()
        t = self.thrust()
        g = 6.674e-11
        r = Planet.distance((0, 0), pos)
        uvec = (-pos[0] / r, -pos[1] / r)
        fg = g * self.mass() * self.planet.mass / r**2
        vf = [x * fg for x in uvec]
        return [vf[0] + t * cos(self.rotation), vf[1] + t * sin(self.rotation)]

    # geric acceleration as a function of position
    def _ar(self, pos):
        """
        Compute the acceleration vector of the rocket, as a function of the
        position vector.
        """
        m = self.mass()
        vf = self._fr(pos)
        return [vf[i] / m for i in (0, 1)]

    # add a status reporting function to status display
    def addStatusReport(self, statuslist, statusfuncs, statusselect):
        """
        Accept list of all status names, all status text functions, and
        the list of status names that have been selected for display.

        :param list[str] statuslist: List of status names to include in flight
            parameters. Default is all, consisting of: "velocity", "acceleration",
            "course", "altitude", "thrust", "mass", "trueanomaly", "scale", "timezoom",
            and "shiptime".
        :param list[func] statusfuncs: List of function references corresponding
            to statuslist parameter.
        :param list[str] statusselect: List of names chosen from statuslist to display.

        :returns: None
        """
        statusdict = dict(zip(statuslist, statusfuncs))
        for name in statusselect:
            if name in statusdict:
                self._labels.append(
                    Label(
                        self._statuspos[:],
                        statusdict[name],
                        size=15,
                        positioning="physical",
                        width=250,
                    )
                )
                self._statuspos[1] += 25

    @property
    def xyposition(self):
        """
        Report the x,y tuple for logical position of the spaceship.
        """
        return self._xy

    @xyposition.setter
    def xyposition(self, pos):
        self._xy = pos
        # self._touchAsset()

    @property
    def tanomalyd(self):
        """
        Report/set the spaceship position as a direction relative to central body
        (degrees).
        """
        return degrees(self.tanomaly)

    @tanomalyd.setter
    def tanomalyd(self, angle):
        self.tanomaly = radians(angle)

    @property
    def altitude(self):
        """
        Report/set the spaceship altitude of planet surface in meters.
        """
        alt = Planet.distance(self._xy, (0, 0)) - self.planet.radius
        return alt

    @altitude.setter
    def altitude(self, alt):
        r = alt + self.planet.radius
        self._xy = (r * cos(self.tanomaly), r * sin(self.tanomaly))

    @property
    def velocity(self):
        """
        Report the spaceship velocity scalar in m/s.
        """
        return self._vmag(self._v_vect)

    @property
    def acceleration(self):
        """
        Report the spaceship acceleration scalar in m/s.
        """
        return self._vmag(self._a_vect)

    @property
    def tanomaly(self):
        """
        Report/set the spaceship position as a direction relative to central body
        (radians).
        """
        # pos = self._pos()
        return atan2(self._xy[1], self._xy[0])

    @tanomaly.setter
    def tanomaly(self, angle):
        r = self.r
        self._xy = (r * cos(angle), r * sin(angle))
        self.touchAsset()

    @property
    def r(self):
        """
        Report the spaceship distance (radius) from central body center of mass.
        """
        return self.altitude + self.planet.radius



"""
ggame package defines names that may be imported directly from ggame (legacy)
"""
from .__version__ import VERSION
from .asset import ImageAsset, TextAsset, CircleAsset, RectangleAsset
from .asset import PolygonAsset, LineAsset, EllipseAsset
from .asset import Frame, Color, LineStyle, BLACK, WHITE, BLACKLINE, WHITELINE
from .sound import SoundAsset, Sound
from .sprite import Sprite
from .app import App
from .event import KeyEvent, MouseEvent
